#!/usr/bin/env python

# author: Youri Lammers
# contact: youri.lammers@uit.no / youri.lammers@gmail.com

# Parse the multiplex results to identify the vaccinium references and
# groups present. The script can rarefy the data as well, though it doesn't
# do this by default

# import the modules used by the script
import argparse, itertools, re, sys, os, random
from collections import defaultdict

# Retrieve the commandline arguments
parser = argparse.ArgumentParser(description = 'Extract the amplicons.')

parser.add_argument('-t','--tables', metavar='OBITab tables', dest='tables',
			action='append', nargs='+',
			help='List of the input OBITab tables.')
parser.add_argument('-f','--fasta', metavar='Fasta files', dest='ff',
			action='append', nargs='+',
			help='List of fasta files')
parser.add_argument('-min_read', metavar='minimum reads', dest='minread',
			default=3, type=int,
			help='The minimum number of reads per barcode and detection (default=3).')
parser.add_argument('-min_rep', metavar='minimum repeats', dest='minrep',
			default=2, type=int,
			help='The minimum number of repeats per barcode (default=2).')
parser.add_argument('-rare_level', metavar='number of rarefaction passes',
			dest='rarelevel', default=1, type=int,
			help='The number of rarefaction rounds (default=1).')
parser.add_argument('-rarefy', metavar='number of rarefied reads',
			dest='rareread', default=100000000, type=int,
			help='The number of reads to rarefy to (default=100000000).')
parser.add_argument('-ts', metavar='tag-swap rate', dest='tsr',
			default=0.01, type=float,
			help='The tag swap rate used for filtering (default=0.01).')

args = parser.parse_args()


def extract_sequences(file):

	# open the sequence file submitted by the user
	sequence_file = open(file)

	# parse throught the file
	lines = (x[1] for x in itertools.groupby(sequence_file,
		key=lambda line: line[0] == '>'))

	# walk through the header and obtain the sequences
	for headers in lines:

		header = headers.next().strip().lower().replace('>>','>').split(';')[0]
		if header == '': continue

		sequence = ''.join(line.strip() for line in lines.next()).lower()

		# yield the header + sequence
		yield [header, sequence]


def parse_table(path, ref_dict, rep_count, sample2rep):

	# get the count, iden and marker data from a library

	# get the sample name
	sample = os.path.splitext(os.path.basename(path))[0].split('.')[0]
	
	# get a list with the excepted sequences
	sequence_lookup = {}

	# populate the sequence list
	for k,v in ref_dict.items():
		for a,b in ref_dict[k].items():
			sequence_lookup[a] = k

	# create the empty variables for storing the column locations
	seqpos, countpos, libsum = 0, 0, 0
	locations = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))

	# go through the file
	for line in open(path):

		# split the lines into the columns
		line = line.strip().split('\t')

		# if the line is the header, extract the column locations
		if line[0] == 'id':

			# loop through the columns
			for i in range(0,len(line)):

				# store the location of the relevant columns
				if line[i] == 'count': countpos = i
				if line[i] == 'sequence': seqpos = i
				if 'sample' in line[i]:

					marker = line[i].replace('sample:','').split('_rpt')[0]
					if len(marker) == 4: marker = marker.replace('Vu_','Vu_0')
					rep = line[i].split('_rpt')[1]
					locations[marker]['data'][rep] = i

				elif 'obiclean_status:' in line[i]:

					marker = line[i].replace('obiclean_status:','').split('_rpt')[0]
					if len(marker) == 4: marker = marker.replace('Vu_','Vu_0')
					rep = line[i].split('_rpt')[1]
					locations[marker]['obiclean'][rep] = i
					
			# continue with the next line after storing the header info
			continue

		# define the sequence
		sequence = line[seqpos]

		# loop through the markers
		for marker in locations.keys():

			# loop through the replicate columns for the marker
			for rep in locations[marker]['data']:

				# get the read count
				readcount = int(line[locations[marker]['data'][rep]])

				# keep track of the total rep count
				rep_count[sample][marker][rep] += readcount

				# continue if the sequence is not in the ref
				if sequence not in sequence_lookup: continue

				# continue if the sequence is identified as internal by obiclean
				obiclean = line[locations[marker]['obiclean'][rep]]
				if obiclean == 'i': continue

				# save the data
				sample2rep[sample][marker][sequence][rep] = readcount

	# return everything
	return [rep_count, sample2rep]


def get_ref():

	# set up the results dictionary
	ref_dict = defaultdict(lambda: defaultdict(list))
		
	# loop through the files
	for file in args.ff[0]:

		# get a read from the input file
		for seq in extract_sequences(file):

			# only use the sequences of interest
			if re.search(r'>vu_[0-9]{2}_vaccinium_uliginosum', seq[0]):

				# get the sample from the fasta hearder by removing
				# the marker and contig info
				sample = re.sub(r'>vu_[0-9]{2}_','',seq[0].split('.')[0])

				# get the marker
				marker = file.split('/')[1].replace('.fasta','')

				# add the sequence to the results
				ref_dict[marker][seq[1]].append(sample)

	# return the ref sequences
	return ref_dict


def calc_prop(count_dict):

	# accept a dictionary with counts and return the proportions

	# calculate the total read count
	total = sum([count_dict[entry] for entry in count_dict])

	# create a new dictionary for the proportions
	prop_dict = {}

	# calculate the proportions
	for entry in count_dict:
		prop_dict[entry] = float(count_dict[entry])/total

	# return the resulting dict
	return prop_dict


def calc_expected(tot_count, prop, rate):

	# calculate and return the swap rate
	return tot_count * prop * rate


def remove_tag_swap(sample2rep, rep_count):

	# function to remove pcr tag swaps at the set rate

	# create a new dict for the filtered results
	new_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))

	# parse through the data and calculate the tag swaps
	for sample in sample2rep:
		for marker in sample2rep[sample]:

			# get the rep proportions for the sample / marker
			rep_prop = calc_prop(rep_count[sample][marker])

			# loop through the sequences and reps present
			for seq in sample2rep[sample][marker]:

				# get the total count across the repeats
				tot_rep = sum([sample2rep[sample][marker][seq][rep] for 
					rep in sample2rep[sample][marker][seq]])

				# calculate the tag swap rates across the replicates
				for rep in sample2rep[sample][marker][seq]:

					# get the count and prop
					count = sample2rep[sample][marker][seq][rep]
					prop = rep_prop[rep]

					# skip if count is zero
					if count == 0: continue

					# calculate the expected read count
					expected = calc_expected(tot_rep, prop, args.tsr)

					# reset if it can be explained by a swap
					if count > expected:
						new_dict[sample][marker][seq][rep] = sample2rep[sample][marker][seq][rep]

	# return the result dictionaries
	return new_dict


def create_seq_lookup(sample2rep):

	# create a dictionary lookup for the sequences
	seq2marker = defaultdict(str)

	# parse through the data
	for sample in sample2rep:
		for marker in sample2rep[sample]:
			for seq in sample2rep[sample][marker]:
				if seq not in seq2marker:
					seq2marker[seq] = marker

	return seq2marker


def remove_low_freq(marker2rep):

	# create a new dict
	new_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))

	# loop through the samples, markers and seqs
	for marker in marker2rep:
		for seq in marker2rep[marker]:

			# remove low read counts
			for rep in marker2rep[marker][seq]:
				if marker2rep[marker][seq][rep] < args.minread:
					marker2rep[marker][seq][rep] = 0

			# get the number of repeats with reads
			totrep = len([rep for rep in marker2rep[marker][seq] if marker2rep[marker][seq][rep] > 0])

			# remove if lower
			if totrep >= args.minrep:
				new_dict[marker][seq] = marker2rep[marker][seq]

	return new_dict



def convert_results_dict(marker2rep, ref_dict):

	# convert the marker2rep dict to include the reference information
	# and add the total rep and count data for easy output

	# set up the new dict
	new_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))

	# parse through the data
	for marker in marker2rep:
		for seq in marker2rep[marker]:

			# get a list of the matching references
			reflist = ref_dict[marker][seq]

			# get the number of replicates and total read count
			totrep = len([rep for rep in marker2rep[marker][seq] if marker2rep[marker][seq][rep] > 0])
			totread = sum([marker2rep[marker][seq][rep] for rep in marker2rep[marker][seq]])

			# loop through the matching references
			# and add the data to the new dict
			for ref in reflist:
				for rep in marker2rep[marker][seq]:
					new_dict[marker][seq][ref][rep] = marker2rep[marker][seq][rep]
			
			# add the total data
			new_dict[marker][seq]['totread'] = totread
			new_dict[marker][seq]['totrep'] = totrep

	# return the new dict
	return new_dict


def rarefy(marker2rep, seq2marker):

	# return a rarefied dictionary

	# create the output dict
	raredict = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))

	# create the lists for rarefying
	seqlist = defaultdict(list)

	for marker in marker2rep:
		for seq in marker2rep[marker]:

			# get the number of reads per repeat and add it to the list
			for rep in marker2rep[marker][seq]:
				seqlist[rep] += ([seq] * marker2rep[marker][seq][rep])

	# shuffle the lists
	for rep in seqlist:
		random.shuffle(seqlist[rep])

	# go through the shuffled lists and pull a subset
	# convert the subset back to the sample2rep format
	for rep in seqlist:

		# create the subset
		if len(seqlist[rep]) > args.rareread:
			subset = seqlist[rep][0:args.rareread]
		else:
			subset = seqlist[rep]

		# loop through the seqs
		for seq in subset:

			# get the marker
			marker = seq2marker[seq]

			# add it to the rare dict
			raredict[marker][seq][rep] += 1

	return raredict


def avg(list):
	# return a rounded average
	if len(list) > 0:
		return round(float(sum(list))/len(list),3)
	else:
		return 0


def get_reference_to_keep(marker2rep, reflist):

	# return a list with reference sequences to keep
	keep = []

	# get a list with the sequences present for the sample
	seqlist = [[seq for seq in marker2rep[marker] if marker2rep[marker][seq]['totrep'] >= args.minrep] for marker in marker2rep]
	seqlist = [b for a in seqlist for b in a]

	# get the markers
	markers = sorted(marker2rep.keys())

	# while there are sequences left for the samples, find the most abundant references(s)
	# that are supported by unique sequences and save them
	while len(seqlist) > 0:

		# keep track of the seqs per ref and the tot seqs and reads per ref
		ref_seqs = defaultdict(list)
		ref_res = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

		# find the reference with the most samples
		for ref in reflist:

			# skip already parsed refs
			if ref in keep: continue

			# keep track of the total reads and markers for the reference
			totseq, totrep, totread = 0, 0, 0

			# loop over the markers and sequences for the sample
			for mark in markers:
				for seq in seqlist:

					# if the reference is present for the sequence
					if ref in marker2rep[mark][seq]:

						# update the seq and read counts and add the seq
						# to the ref set
						totseq += 1
						totread += marker2rep[mark][seq]['totread']
						totrep += marker2rep[mark][seq]['totrep']
						ref_seqs[ref].append(seq)

			# add the ref with the totseq and totread to the result dict
			ref_res[totseq][totrep][totread].append(ref)

		# get the most abundant ref(s)
		maxseq = sorted(ref_res.keys(),reverse=True)[0]
		maxrep = sorted(ref_res[maxseq].keys(),reverse=True)[0]
		maxread = sorted(ref_res[maxseq][maxrep].keys(),reverse=True)[0]
		currefs = ref_res[maxseq][maxrep][maxread]

		#print currefs
		tempremove = []

		# add the refs to the keep list and save the sums data
		for cur in currefs:
			keep.append(cur)
			for seq in ref_seqs[cur]:
				tempremove.append(seq)

		# remove the sequences from the current set from the seq list
		seqlist = [seq for seq in seqlist if seq not in tempremove]

	return keep


def cleanup(marker2rep, keep):

	# remove non-keep reference from the marker2rep set
	clean_dic = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))

	# parse through the data
	for marker in marker2rep:
		for seq in marker2rep[marker]:

			# store the temp data
			tempdata = {}
			numref = 0
			repreads = {}

			# copy the ref data into the new dict if its in the keep list
			for ref in marker2rep[marker][seq]:
				
				# skip for the total count
				if ref == 'total': continue
				
				# keep track of the references and count that are present
				if ref in keep:
					tempdata[ref] = marker2rep[marker][seq][ref]
					numref += 1
					if repreads == {}:
						repreads = marker2rep[marker][seq][ref]

			# divide the rep counts
			for ref in tempdata:
				for rep in tempdata[ref]:
					clean_dic[marker][seq][ref][rep] = round(float(repreads[rep])/numref,3)

	# return the cleaned data
	return clean_dic


def get_group_information(marker2rep, reflist):

	# calculate the number of groups present based on the unique reads
	groupdict = defaultdict(int)

	# create a dict that links each ref to the vaccinium group
	groups = {'vaccinium_uliginosum_canada_trom_v_355000_cdm_apu':'Arctic','vaccinium_uliginosum_russia_trom_v_355008_cdm_aza':'Arctic',
				'vaccinium_uliginosum_france_bgn_afb':'A-A','vaccinium_uliginosum_france_bgn_eap':'Alpine',
				'vaccinium_uliginosum_france_trom_v_355006_cdm_auo':'Alpine','vaccinium_uliginosum_france_trom_v_355013_cdm_azf':'Alpine',
				'vaccinium_uliginosum_russia_trom_v_355010_cdm_azc':'Beringian','vaccinium_uliginosum_microphyllum_alta_trom_v_162265_bxa_bns':'A-A',
				'vaccinium_uliginosum_nordkapp_trom_v_355015_cdm_azg':'Arctic','vaccinium_uliginosum_northkorea_trom_v_355004_cdm_apw':'Arctic',
				'vaccinium_uliginosum_canada_trom_v_355009_cdm_azb':'Arctic','vaccinium_uliginosum_senja_trom_v_963388_bxa_bnr':'A-A',
				'vaccinium_uliginosum_spain_trom_v_355001_cdm_apa':'Alpine','vaccinium_uliginosum_spain_trom_v_355005_cdm_apx':'Alpine',
				'vaccinium_uliginosum_greenland_trom_v_355011_cdm_azd':'Arctic','vaccinium_uliginosum_switzerland_trom_v_355003_cdm_apv':'A-A',
				'vaccinium_uliginosum_uliginosum_loppa_trom_v_962025_bmb_fao':'A-A','vaccinium_uliginosum_svalbard_trom_v_355012_cdm_aze':'Arctic'}

	# loop through the markers present
	for marker in marker2rep:
		for seq in marker2rep[marker]:
			curr_groups = []
			tempdict = defaultdict(int)
			for ref in reflist:
				for rep in marker2rep[marker][seq][ref]:
					# identify the group for each reference present and keep
					# track of the unique number of groups present for a sample
					if marker2rep[marker][seq][ref][rep] > 0:
						tempdict[groups[ref]] += marker2rep[marker][seq][ref][rep]
					if groups[ref] not in curr_groups: curr_groups.append(groups[ref])
			if len(curr_groups) == 1:
				groupdict[curr_groups[0]] += tempdict[curr_groups[0]]
	
	# return the groups present
	return groupdict


def get_rarefied_sample2rep(sample2rep, seq2marker, ref_dict, reflist):

	# run the rarefaction args.rare times and return the average sample2rep dict
	sup5dict = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
	sup6dict = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(float)))))
	storage5dict = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
	storage6dict = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list)))))
	storage7dict = defaultdict(lambda: defaultdict(list))

	for sample in sample2rep:
		#print sample

		# rarefy the data x times
		for i in range(0,args.rarelevel):

			# get the rarefied data
			raredict = rarefy(sample2rep[sample],seq2marker)

			# remove low frequency observations
			raredict = remove_low_freq(raredict)

			# add the reference data
			exp_raredict = convert_results_dict(raredict, ref_dict)

			# work out which references to keep
			keep = get_reference_to_keep(exp_raredict, reflist)

			# save the refs present, and the marker / read data
			exp_raredict = cleanup(exp_raredict, keep)

			# get the group information
			groupdict = get_group_information(exp_raredict, reflist)

			# keep track of the total reads and repeats per ref, before storing
			tempdict = defaultdict(lambda: defaultdict(list))

			# summarize the output
			for marker in exp_raredict:
				for seq in exp_raredict[marker]:
					for ref in reflist:

						# add it to the storage dict for supp table 6
						storage6dict[sample][marker][seq][ref]['read'].append(sum(
							[exp_raredict[marker][seq][ref][rep] for rep in exp_raredict[marker][seq][ref]]))
						storage6dict[sample][marker][seq][ref]['rep'].append(len(
							[rep for rep in exp_raredict[marker][seq][ref] if exp_raredict[marker][seq][ref][rep] > 0]))

						# add it to the temp dict for the the supp 5 average calculation						
						tempdict[ref]['read'].append(sum(
							[exp_raredict[marker][seq][ref][rep] for rep in exp_raredict[marker][seq][ref]]))
						tempdict[ref]['rep'].append(len(
							[rep for rep in exp_raredict[marker][seq][ref] if exp_raredict[marker][seq][ref][rep] > 0]))

			# store the results
			for ref in reflist:
				storage5dict[ref][sample]['read'].append(sum(tempdict[ref]['read']))
				storage5dict[ref][sample]['rep'].append(sum(tempdict[ref]['rep']))
		
			# store the group results
			storage7dict[sample]['groups'].append(len(groupdict))
			storage7dict[sample]['refs'].append(len([tempdict[ref] for ref in tempdict if sum(tempdict[ref]['read']) > 0]))
			sumuniq = sum([groupdict[group] for group in groupdict])
			storage7dict[sample]['maingroup'].append(len([groupdict[group] for group in groupdict if (groupdict[group]/sumuniq)>0.01]))
			for group in ['Arctic','Alpine','Beringian','A-A']:
				storage7dict[sample][group].append(groupdict[group])

		# get the average for each detection and add it to the result dictionaries
		for ref in reflist:

			# calculate the average and add it to the result dict
			sup5dict[ref][sample]['read'] = avg(storage5dict[ref][sample]['read'])
			sup5dict[ref][sample]['rep'] = avg(storage5dict[ref][sample]['rep'])

	# get the average values for supp table 6 and 7
	for sample in storage6dict:
		for marker in storage6dict[sample]:
			for seq in storage6dict[sample][marker]:
				for ref in reflist:
					sup6dict[sample][marker][seq][ref]['read'] = avg(storage6dict[sample][marker][seq][ref]['read'])
					sup6dict[sample][marker][seq][ref]['rep'] = avg(storage6dict[sample][marker][seq][ref]['rep'])

		# add the group info
		sup6dict[sample]['groups'] = avg(storage7dict[sample]['groups'])
		sup6dict[sample]['maingroup'] = avg(storage7dict[sample]['maingroup'])
		sup6dict[sample]['refs'] = avg(storage7dict[sample]['refs'])
		for group in ['Arctic','Alpine','Beringian','A-A']:
			sup6dict[sample][group] = avg(storage7dict[sample][group])

	return [sup5dict, sup6dict]


def analyse_data(sample2rep, ref_dict):

	# function for the parsing and output of the results

	# get the available reference genomes
	reflist = [[[ref for ref in ref_dict[marker][sequence]] 
		for sequence in ref_dict[marker]] for marker in ref_dict]
	reflist = [c for a in reflist for b in a for c in b]
	reflist = sorted(list(set(reflist)))

	# get the samples
	markers, samples = sorted(ref_dict.keys()), sorted(sample2rep.keys())

	# get the seq lookup table
	seq2marker = create_seq_lookup(sample2rep)

	# run the rarefying analysis
	results = get_rarefied_sample2rep(sample2rep, seq2marker, ref_dict, reflist)
	sup5dict = results[0]
	sup6dict = results[1]


	# print the number of groups and references present
	runname = '{0}-{1}'.format(args.minrep,args.minread)
	print 'type\treads\trarefactionruns\tfilter\t{0}'.format('\t'.join(samples))
	print 'groups\t{0}\t{1}\t{2}\t{3}'.format(args.rareread,args.rarelevel,runname,'\t'.join([str(sup6dict[sample]['groups']) for sample in samples]))
	print 'maingroup\t{0}\t{1}\t{2}\t{3}'.format(args.rareread,args.rarelevel,runname,'\t'.join([str(sup6dict[sample]['maingroup']) for sample in samples]))
	print 'references\t{0}\t{1}\t{2}\t{3}'.format(args.rareread,args.rarelevel,runname,'\t'.join([str(sup6dict[sample]['refs']) for sample in samples]))
	print '\n\n\n'

	# format and print the output for supplementary table 5
	print 'reference\t{0}'.format('\t'.join(samples))
	for ref in reflist:
		print '{0}\t{1}'.format(ref,'\t'.join([str(sup5dict[ref][sample]['read']) for sample in samples]))	
	print '\n'
	print 'reference\t{0}'.format('\t'.join(samples))
	for ref in reflist:
		print '{0}\t{1}'.format(ref,'\t'.join([str(sup5dict[ref][sample]['rep']) for sample in samples]))	
	print '\n\n\n'

	# format and print the output for supplementary table 6
	print 'sample\tmarker\tseq\ttype\taverage\tmax\t{0}'.format('\t'.join(reflist))
	for sample in samples:
		for marker in sup6dict[sample]:
			# skip for group info
			if marker in ['Arctic','Alpine','Beringian','A-A','groups', 'refs', 'maingroup']: continue
			for seq in sup6dict[sample][marker]:
				print '{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}'.format(sample,marker,seq,'read',
					str(avg([sup6dict[sample][marker][seq][ref]['read'] for ref in reflist])),
					str(max([sup6dict[sample][marker][seq][ref]['read'] for ref in reflist])),
					'\t'.join([str(sup6dict[sample][marker][seq][ref]['read']) for ref in reflist]))
				print '{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}'.format(sample,marker,seq,'rep',
					str(avg([sup6dict[sample][marker][seq][ref]['rep'] for ref in reflist])),
					str(max([sup6dict[sample][marker][seq][ref]['rep'] for ref in reflist])),
					'\t'.join([str(sup6dict[sample][marker][seq][ref]['rep']) for ref in reflist]))


def main ():

	# get the ref sequences
	ref_dict = get_ref()

	# set up the results dictionary and the min max lenght variables
	rep_count = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
	sample2rep = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int))))
		
	# loop through the samples
	for table in args.tables[0]:

		# get the data from the table
		rep_count, sample2rep = parse_table(
			table, ref_dict, rep_count, sample2rep)

	# identify and remove the tag swaps
	sample2rep = remove_tag_swap(sample2rep, rep_count)

	# analyse and output the data
	analyse_data(sample2rep, ref_dict)


if __name__ == '__main__':
	main()